@@ -1,4 +1,4 @@
-class NewMetric: # discuss the case when more than 1 new metric is defined, new class to be created or design to be in dict form
+class NewMetric:
     """
     Base class to add new metrics to the Veritas library.
 
@@ -16,20 +16,37 @@ class NewMetric: # discuss the case when more than 1 new metric is defined, new
     metric_definition: str
         Full name of metric
 
-    metric_parity_ratio: str
-        "parity" or "ratio"
+    metric_short_name: str
+        Short-form name of metric
+
+    metric_difference_ratio: str
+        "difference" or "ratio". For fairness metric_type only.
+
+    metric_equiv_perf_metric: str
+        Equivalent performance metric. For fairness metric_type only.
+
+    metric_direction: str
+        Direction of performance metric, i.e., whether a `higher` metric value indicates better model performance. For fairness metric_type only.
+        "higher" or "lower"
+
+    metric_reqt: str
+        Metric requirement/dependency, i.e., "y_pred" or "y_prob"
 
     enable_flag: boolean
         Whether the new metric can be a primary metric
     """
     metric_type = "fair"
     metric_group = "classification"
-    metric_name = "custom base"
-    metric_definition = "custom_new_metric base"
-    metric_parity_ratio = "parity"
+    metric_name = "custom_base"
+    metric_definition = "custom new metric base"
+    metric_short_name = "custom"
+    metric_difference_ratio = "difference"
+    metric_equiv_perf_metric = 'balanced_acc'
+    metric_direction = 'higher'
+    metric_reqt = 'y_prob'
     enable_flag = True
 
-    def compute(self):
+    def compute(self, **kwargs):
         """
         Returns tuple of metric value and privileged group value
 
