@@ -61,7 +61,7 @@ class TradeoffRate(object):
         map_metric_to_method : dict
                 Mapping of metric name to respective compute function
                 
-        sigma : float or integer , default = 0
+        sigma : float or integer, default = 0
                  Standard deviation for Gaussian kernel for smoothing the contour lines of primary fairness metric. 
                  When sigma <= 0, smoothing is turn off.
                  Suggested to try sigma = 3 or above if noisy contours are observed.
@@ -70,6 +70,8 @@ class TradeoffRate(object):
         self.perf_metric_name = usecase_obj.perf_metric_name
         self.fair_metric_name = usecase_obj.fair_metric_name
         self.metric_group = FairnessMetrics.map_fair_metric_to_group[self.fair_metric_name][1]
+        # replacement flag to indicate if fair_metric_name has been replaced with another metric from fairness tree
+        self.replacement_flag = 0
         if self.metric_group == "uplift":
             if usecase_obj.spl_params['revenue'] is not None and usecase_obj.spl_params['treatment_cost'] is not None:                
                 if self.perf_metric_name not in ["expected_profit", "emp_lift"]:
@@ -84,7 +86,8 @@ class TradeoffRate(object):
                 self.perf_metric_name = "balanced_acc"
             is_valid_metric = FairnessMetrics.map_fair_metric_to_group[self.fair_metric_name][3]
             if is_valid_metric == False:
-                self.fair_metric_name = usecase_obj._select_fairness_metric_name() #usecase._select_fairness_metric_name() get the usecase object over
+                self.fair_metric_name = usecase_obj._fairness_tree(is_pos_label_favourable = usecase_obj.fair_is_pos_label_fav) #usecase._fairness_tree to get replacement metric
+                self.replacement_flag = 1 #set replacement flag to 1 if replacement metric is used in tradeoff
 
         self.p_var = usecase_obj.model_params[0].p_var
         self.curr_p_var = None
@@ -122,11 +125,23 @@ class TradeoffRate(object):
         'equal_odds': self._compute_equalized_odds_tr,
         'neg_equal_odds': self._compute_negative_equalized_odds_tr,
         'calibration_by_group': self._compute_calibration_by_group_tr,
+        'equal_opportunity_ratio': self._compute_equal_opportunity_ratio_tr,
+        'for_ratio': self._compute_false_omission_rate_ratio_tr,
+        'fdr_ratio': self._compute_false_discovery_rate_ratio_tr,
+        'ppv_ratio': self._compute_positive_predictive_ratio_tr,
+        'npv_ratio': self._compute_negative_predictive_ratio_tr,
+        'tnr_ratio': self._compute_tnr_ratio_tr,
+        'fnr_ratio': self._compute_fnr_ratio_tr,
+        'fpr_ratio': self._compute_fpr_ratio_tr,
+        'equal_odds_ratio': self._compute_equalized_odds_ratio_tr,
+        'neg_equal_odds_ratio': self._compute_negative_equalized_odds_ratio_tr,
+        'calibration_by_group_ratio': self._compute_calibration_by_group_ratio_tr,
         'rejected_harm': self._compute_rejected_harm_tr,
         'expected_profit': self._compute_expected_profit_tr,
         'emp_lift': self._compute_emp_lift_tr
         }
         self.sigma = usecase_obj.sigma
+
     def compute_tradeoff(self, n_threads, tdff_pbar):
         """
         Computes the tradeoff values.
@@ -171,8 +186,12 @@ class TradeoffRate(object):
             ## iterate through the protected_features
             for i in self.p_var:
                 self.curr_p_var = i
-                self.uplift_rates_a = ModelRateUplift(self.y_true, self.pred_outcome, self.e_lift, self.feature_mask[i], cost, revenue, self.proportion_of_interpolation_fitting, n_threads) 
-                self.uplift_rates_b = ModelRateUplift(self.y_true, self.pred_outcome, self.e_lift, ~self.feature_mask[i], cost, revenue, self.proportion_of_interpolation_fitting, n_threads)
+                mask = self.feature_mask[i]
+                mask_p = mask==1                
+                mask_up = mask==0
+                
+                self.uplift_rates_a = ModelRateUplift(self.y_true, self.pred_outcome, self.e_lift, mask_p, cost, revenue, self.proportion_of_interpolation_fitting, n_threads) 
+                self.uplift_rates_b = ModelRateUplift(self.y_true, self.pred_outcome, self.e_lift, mask_up, cost, revenue, self.proportion_of_interpolation_fitting, n_threads)
                 #### compute the metrics and create dictionary
                 fair_values = self.map_metric_to_method[self.fair_metric_name]()
                 perf_values = self.map_metric_to_method[self.perf_metric_name]()
@@ -209,15 +228,20 @@ class TradeoffRate(object):
             prog = round(80/(len(self.p_var)), 2)
 
             # iterate through the protected_features
-            for i in self.p_var:
+            for idx, i in enumerate(self.p_var):
+                
                 self.curr_p_var = i
+                mask = self.feature_mask[i]
+                mask_p = mask==1                
+                mask_up = mask==0
+                
                 # initialising ModelRates object to access interpolated base rates
                 if self.sample_weight is None :
-                    self.rates_a = ModelRateClassify(self.y_true[self.feature_mask[i]], self.y_prob[self.feature_mask[i]])
-                    self.rates_b = ModelRateClassify(self.y_true[~self.feature_mask[i]], self.y_prob[~self.feature_mask[i]])                
+                    self.rates_a = ModelRateClassify(self.y_true[mask_p], self.y_prob[mask_p])
+                    self.rates_b = ModelRateClassify(self.y_true[mask_up], self.y_prob[mask_up])                             
                 else :
-                    self.rates_a = ModelRateClassify(self.y_true[self.feature_mask[i]], self.y_prob[self.feature_mask[i]], self.sample_weight[self.feature_mask[i]])
-                    self.rates_b = ModelRateClassify(self.y_true[~self.feature_mask[i]], self.y_prob[~self.feature_mask[i]], self.sample_weight[~self.feature_mask[i]])
+                    self.rates_a = ModelRateClassify(self.y_true[mask_p], self.y_prob[mask_p], self.sample_weight[mask_p])
+                    self.rates_b = ModelRateClassify(self.y_true[mask_up], self.y_prob[mask_up], self.sample_weight[mask_up])
 
                 # base rates for privileged and unprivileged groups to be used in classification based compute functions
                 self.tpr_a, self.tpr_b, self.fpr_a, self.fpr_b, self.selection_rate_a, self.selection_rate_b, self.forr_a, self.forr_b, self.ppv_a, self.ppv_b = self.rates_a.tpr(self.th_a), self.rates_b.tpr(self.th_b), self.rates_a.fpr(self.th_a), self.rates_b.fpr(self.th_b), self.rates_a.selection_rate(self.th_a), self.rates_b.selection_rate(self.th_b), self.rates_a.forr(self.th_a), self.rates_b.forr(self.th_b), self.rates_a.ppv(self.th_a), self.rates_b.ppv(self.th_b)
@@ -235,21 +259,26 @@ class TradeoffRate(object):
 
                 fair_values_smooted = gaussian_filter(fair_values, self.sigma)
                 self.result[i]['fair'] = fair_values_smooted
-                
+
                 self.result[i]['perf'] = perf_values
 
                 self.result[i]['th_x'] = self.th_x
                 self.result[i]['th_y'] = self.th_y
 
                 best_th1, best_th2, best_th3 = TradeoffRate._compute_max_perf(self,perf_values,fair_values)
+                
+                # Store threshold and perf metric value for single_th
+                if idx==0:
+                    best_th1_single_th = best_th1
+
                 self.result[i]['max_perf_point'] = best_th2
-                self.result[i]['max_perf_single_th'] = best_th1
+                self.result[i]['max_perf_single_th'] = best_th1_single_th
                 self.result[i]['max_perf_neutral_fair'] = best_th3
                     
                 tdff_pbar.update(prog)
         else :
             self.result = None
-            self.msg = "Tradeoff has been skipped due to fair_metric_name"
+            self.msg = "Tradeoff is skipped due to fair_metric_name not supported"
             return
 
     @staticmethod
@@ -317,13 +346,20 @@ class TradeoffRate(object):
         tpr_a, tpr_b = self.rates_a.tpr(self.th_a), self.rates_b.tpr(self.th_b)
         fpr_a, fpr_b = self.rates_a.fpr(self.th_a), self.rates_b.fpr(self.th_b)
         mask = self.feature_mask[self.curr_p_var]
+        maskFilterNeg = mask==-1
+        mask_ma = np.ma.array(mask, mask = maskFilterNeg)
+
         # Performance
         # Combine TPRs: P(R=1|Y=1) = P(R=1|Y=1,A=1)P(A=1|Y=1) + P(R=1|Y=1,A=0)P(A=0|Y=1)
-        tpr = (tpr_a * np.mean(mask[self.y_true == 1]) +
-               tpr_b * np.mean(~mask[self.y_true == 1])) ####self.y_true[0]
+        # tpr = (tpr_a * np.mean(mask[self.y_true == 1]) +
+        #        tpr_b * np.mean(~mask[self.y_true == 1])) ####self.y_true[0]
+        tpr = (tpr_a * np.mean(mask_ma[self.y_true == 1].astype(bool)) +
+                tpr_b * np.mean(~mask_ma[self.y_true == 1].astype(bool)))
         # Combine FPRs: P(R=1|Y=0) = P(R=1|Y=0,A=1)P(A=1|Y=0) + P(R=1|Y=0,A=0)P(A=0|Y=0)
-        fpr = (fpr_a * np.mean(mask[self.y_true == 0]) +
-               fpr_b * np.mean(~mask[self.y_true == 0]))
+        # fpr = (fpr_a * np.mean(mask[self.y_true == 0]) +
+        #        fpr_b * np.mean(~mask[self.y_true == 0]))
+        fpr = (fpr_a * np.mean(mask_ma[self.y_true == 0].astype(bool)) +
+               fpr_b * np.mean(~mask_ma[self.y_true == 0].astype(bool)))
         bal_acc = 0.5 * (tpr + 1 - fpr)
         return bal_acc
 
@@ -340,13 +376,15 @@ class TradeoffRate(object):
         ppv_a, ppv_b = self.rates_a.ppv(self.th_a), self.rates_b.ppv(self.th_b)
         
         mask = self.feature_mask[self.curr_p_var]
+        maskFilterNeg = mask==-1
+        mask_ma = np.ma.array(mask, mask = maskFilterNeg)
         # Performance
-        # Combine TPRs: P(R=1|Y=1) = P(R=1|Y=1,A=1)P(A=1|Y=1) + P(R=1|Y=1,A=0)P(A=0|Y=1)
-        tpr = (tpr_a * np.mean(mask[self.y_true == 1]) +
-               tpr_b * np.mean(~mask[self.y_true == 1]))
-        # Combine FPRs: P(R=1|Y=0) = P(R=1|Y=0,A=1)P(A=1|Y=0) + P(R=1|Y=0,A=0)P(A=0|Y=0)
-        ppv = (ppv_a * [np.mean(mask[self.y_prob > th]) for th in self.th_a.ravel()] ) + \
-                (ppv_b * [np.mean(~mask[self.y_prob >th]) for th in self.th_b.ravel()] )
+        # Combine TPRs: P(R=1|Y=1) = P(R=1|Y=1,A=1)P(A=1|Y=1) + P(R=1|Y=1,A=0)P(A=0|Y=1)        
+        tpr = (tpr_a * np.mean(mask_ma[self.y_true == 1].astype(bool)) +
+               tpr_b * np.mean(~mask_ma[self.y_true == 1].astype(bool)))
+        # Combine FPRs: P(R=1|Y=0) = P(R=1|Y=0,A=1)P(A=1|Y=0) + P(R=1|Y=0,A=0)P(A=0|Y=0)        
+        ppv = (ppv_a * [np.mean(mask_ma[self.y_prob > th].astype(bool)) for th in self.th_a.ravel()] ) + \
+                (ppv_b * [np.mean(~mask_ma[self.y_prob >th].astype(bool)) for th in self.th_b.ravel()] )
         
         f1 = 2 * ((ppv*tpr)/(ppv+tpr))
         return f1
@@ -362,6 +400,17 @@ class TradeoffRate(object):
         """
         return self.tpr_a - self.tpr_b
 
+    def _compute_equal_opportunity_ratio_tr(self):
+        """
+        Computes the ratio of equal opportunity
+
+        Returns
+        ---------
+        _compute_equal_opportunity_ratio_tr : float
+                tradeoff value
+        """
+        return self.tpr_b/self.tpr_a 
+
     def _compute_disparate_impact_tr(self):
         """
         Computes the ratio of approval rate between the privileged and unprivileged groups
@@ -371,7 +420,7 @@ class TradeoffRate(object):
         _compute_disparate_impact_tr : float
                 tradeoff value
         """
-        return self.selection_rate_a / self.selection_rate_b
+        return self.selection_rate_b/ self.selection_rate_a 
 
     def _compute_demographic_parity_tr(self):
         """
@@ -386,7 +435,7 @@ class TradeoffRate(object):
 
     def _compute_false_omission_rate_parity_tr(self):
         """
-        Computes the difference in negative predictive values between the privileged and unprivileged groups
+        Computes the ratio of false omission rate values between the privileged and unprivileged groups
 
         Returns
         ---------
@@ -395,6 +444,17 @@ class TradeoffRate(object):
         """
         return self.forr_a - self.forr_b
 
+    def _compute_false_omission_rate_ratio_tr(self):
+        """
+        Computes the ratio of false omission rate values between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_false_omission_rate_ratio_tr : float
+                tradeoff value
+        """
+        return self.forr_b/ self.forr_a  
+
     def _compute_false_discovery_rate_parity_tr(self):
         """
         Computes the difference in false discovery rate values between the privileged and unprivileged groups
@@ -406,6 +466,17 @@ class TradeoffRate(object):
         """
         return self.ppv_b - self.ppv_a
 
+    def _compute_false_discovery_rate_ratio_tr(self):
+        """
+        Computes the ratio of false discovery rate values between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_false_discovery_rate_ratio_tr : float
+                tradeoff value
+        """
+        return self.ppv_b / self.ppv_a
+
     def _compute_positive_predictive_parity_tr(self):
         """
         Computes the difference in positive predictive values between the privileged and unprivileged groups
@@ -417,16 +488,39 @@ class TradeoffRate(object):
         """
         return self.ppv_a - self.ppv_b
 
+    def _compute_positive_predictive_ratio_tr(self):
+        """
+        Computes the ratio of positive predictive values between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_positive_predictive_ratio_tr : float
+                tradeoff value
+        """
+        return self.ppv_b/self.ppv_a
+
     def _compute_negative_predictive_parity_tr(self):
         """
         Computes the difference in negative predictive values between the privileged and unprivileged groups
         
         Returns
         ---------
-        None
+        _compute_negative_predictive_parity_tr : float
+                tradeoff value
         """
         return self.forr_b - self.forr_a
 
+    def _compute_negative_predictive_ratio_tr(self):
+        """
+        Computes the ratio of negative predictive values between the privileged and unprivileged groups
+        
+        Returns
+        ---------
+        _compute_negative_predictive_ratio_tr : float
+                tradeoff value
+        """
+        return self.forr_b / self.forr_a
+
     def _compute_tnr_parity_tr(self):
         """
         Computes the difference in true negative rates between the privileged and unprivileged groups
@@ -438,6 +532,17 @@ class TradeoffRate(object):
         """
         return self.fpr_b - self.fpr_a
 
+    def _compute_tnr_ratio_tr(self):
+        """
+        Computes the ratio of true negative rates between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_fnr_ratio_tr : float
+                tradeoff value
+        """
+        return self.fpr_b / self.fpr_a
+
     def _compute_fnr_parity_tr(self):
         """
         Computes the difference in false negative rates between the privileged and unprivileged groups
@@ -449,6 +554,17 @@ class TradeoffRate(object):
         """
         return self.tpr_b - self.tpr_a
 
+    def _compute_fnr_ratio_tr(self):
+        """
+        Computes the ratio of false negative rates between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_fnr_ratio_tr : float
+                tradeoff value
+        """
+        return self.tpr_b / self.tpr_a
+
     def _compute_fpr_parity_tr(self):
         """
         Computes the difference in false positive rates between the privileged and unprivileged groups
@@ -460,6 +576,17 @@ class TradeoffRate(object):
         """
         return self.fpr_a - self.fpr_b
 
+    def _compute_fpr_ratio_tr(self):
+        """
+        Computes the ratio of false positive rates between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_fpr_ratio_tr : float
+                tradeoff value
+        """
+        return self.fpr_b/ self.fpr_a 
+
     def _compute_equalized_odds_tr(self):
         """
         Computes the difference in equalized odds between the privileged and unprivileged groups 
@@ -469,7 +596,19 @@ class TradeoffRate(object):
         _compute_equalized_odds_tr : float
                 tradeoff value
         """
-        return ((self.tpr_a - self.tpr_b) + (self.fpr_a - self.fpr_b))/2
+        return ((self.tpr_a + self.fpr_a) - (self.tpr_b + self.fpr_b))/2
+
+    def _compute_equalized_odds_ratio_tr(self):
+        """
+        Computes the ratio of equalized odds between the privileged and unprivileged groups 
+
+        Returns
+        ---------
+        _compute_equalized_odds_ratio_tr : float
+                tradeoff value
+        """
+        return ((self.tpr_b + self.fpr_b) / (self.tpr_a + self.fpr_a))/2
+
 
     def _compute_negative_equalized_odds_tr(self):
         """
@@ -480,7 +619,19 @@ class TradeoffRate(object):
         _compute_negative_equalized_odds_tr : float
                 tradeoff value
         """
-        return ((self.fpr_b - self.fpr_a) + (self.tpr_b - self.tpr_a))/2 
+        return ((self.fpr_b + self.tpr_b ) - ( self.fpr_a + self.tpr_a))/2
+
+    def _compute_negative_equalized_odds_ratio_tr(self):
+        """
+        Computes the ratio of negative equalized odds between the privileged and unprivileged groups 
+
+        Returns
+        ---------
+        _compute_negative_equalized_odds_ratio_tr : float
+                tradeoff value
+        """
+        return ((self.fpr_b + self.tpr_b - 2) / ( self.fpr_a + self.tpr_a - 2))/2
+
 
     def _compute_calibration_by_group_tr(self):
         """
@@ -491,7 +642,18 @@ class TradeoffRate(object):
         _compute_calibration_by_group_tr : float
                 tradeoff value
         """
-        return ((self.ppv_a - self.ppv_b) + (self.forr_a - self.forr_b)) / 2
+        return ((self.ppv_a + self.forr_a) - (self.ppv_b + self.forr_b)) / 2
+
+    def _compute_calibration_by_group_ratio_tr(self):
+        """
+        Computes the ratio of calibration by group between the privileged and unprivileged groups
+
+        Returns
+        ---------
+        _compute_calibration_by_group_ratio_tr : float
+                tradeoff value
+        """
+        return ((self.ppv_b + self.forr_b) / (self.ppv_a + self.forr_a)) / 2
 
     def _compute_rejected_harm_tr(self):
         """
@@ -524,7 +686,9 @@ class TradeoffRate(object):
         _compute_emp_lift_tr : float
                 tradeoff value
         """
-        mask = self.feature_mask[self.curr_p_var]
+        mask = self.feature_mask[self.curr_p_var]        
+        maskFilterNeg = mask==-1
+        mask = np.ma.array(mask, mask = maskFilterNeg).astype(bool)
         y_true_a = self.y_true[1][mask]
         tr_p = np.array([sum(y_true_a[self.e_lift[mask]> th] =='TR') for th in self.th_a.ravel()])
         tn_p = np.array([sum(y_true_a[self.e_lift[mask]> th] =='TN') for th in self.th_a.ravel()])
